from collections.abc import Callable

from python_ruler.library import compute


def length[T](argument: Callable[[T], list] | list) -> Callable[[T], int]:
    def inner(context: T) -> int:
        return len(compute(argument, context))

    return inner



import importlib
from collections.abc import Callable
from typing import Any, List

from hy.models import Expression, Float, Integer, Object, String, Symbol

from .lexer import Symbol as genSymbol


def compute[T, U, V](argument: Callable[[T], U] | V, context: T) -> U | V:
    return (
        argument(context)  # type: ignore
        if callable(argument)
        else argument  # type: ignore
    )


def evaluate(sequence: Expression, result=None) -> tuple[Any] | Callable[[Any], Any]:
    assert isinstance(sequence, Expression)

    result = result or tuple()
    to_return = None

    if len(sequence) > 0:
        if isinstance(sequence[0], Expression) and sequence[0][0] == Symbol("."):
            to_return = evaluate(
                sequence[1:],  # type: ignore
                result
                + (
                    get_function(
                        f"genruler.modules.{sequence[0][1]}", str(sequence[0][2])
                    ),
                ),
            )

        elif isinstance(sequence[0], Expression):
            to_return = evaluate(
                sequence[1:],  # type: ignore
                result + (evaluate(sequence[0]),),
            )

        else:
            to_return = evaluate(
                sequence[1:],  # type: ignore
                result + (extract(sequence[0]),),
            )

    else:
        if callable(result[0]):
            to_return = result[0](*result[1:])
        else:
            to_return = result

    return to_return


def evaluate2(sequence: List[Any], result=None) -> tuple[Any] | Callable[[Any], Any]:
    """Evaluate a parsed S-expression tree into a callable function.

    Args:
        sequence: The parsed expression tree, where each node is a list:
            - First element is a Symbol object for function names (e.g. Symbol('boolean.tautology'))
            - Remaining elements are arguments, which can be subtrees

    Returns:
        A callable function that takes a context argument and returns a value

    Raises:
        ValueError: If the expression is invalid or empty
        TypeError: If the evaluated expression is not callable
    """
    result = result or tuple()
    to_return = None

    if len(sequence) > 0:
        if isinstance(sequence[0], genSymbol):
            assert "." in sequence[0].name
            module, function = sequence[0].name.split(".")
            to_return = evaluate2(
                sequence[1:],  # type: ignore
                result + (get_function(f"genruler.modules.{module}", function),),
            )

        elif isinstance(sequence[0], list):
            to_return = evaluate2(
                sequence[1:],  # type: ignore
                result + (evaluate2(sequence[0]),),
            )

        else:
            to_return = evaluate2(
                sequence[1:],  # type: ignore
                result + (sequence[0],),
            )

    else:
        if callable(result[0]):
            to_return = result[0](*result[1:])
        else:
            to_return = result

    return to_return


def extract(argument: type[Object]) -> Any:
    match argument.__class__.__name__:
        case String.__name__:
            return str(argument)
        case Integer.__name__:
            return int(argument)
        case Float.__name__:
            return float(argument)
        case _:
            return str(argument)


def get_function(module_name: str, function_name: str) -> Callable[[Any], Any]:
    """Get a function from a module by name."""
    module = importlib.import_module(module_name)

    try:
        function = getattr(module, function_name)
    except AttributeError:
        function = getattr(module, f"{function_name}_")

    assert callable(function)

    return function


def compute(value: Any, context: Any) -> Any:
    """Compute a value in a context."""
    if callable(value):
        return value(context)

    return value



import operator
from collections.abc import Callable
from typing import Any

from genruler.common import binary
from genruler.library import compute


def equal(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.eq, arguments)


def gt(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.gt, arguments)


def ge(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.ge, arguments)


def in_(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.contains, arguments[::-1])


def is_none(argument: Any) -> Callable[[dict[Any, Any]], bool]:
    def inner(context: dict[Any, Any]) -> bool:
        return compute(argument, context) is None

    return inner

def is_true(argument: Any) -> Callable[[dict[Any, Any]], bool]:
    def inner(context: dict[Any, Any]) -> bool:
        return compute(argument, context) is True

    return inner


def lt(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.lt, arguments)


def le(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.le, arguments)



from collections.abc import Callable
from operator import itemgetter
from typing import Any

from genruler.library import compute


def context(context_sub: Any, argument: Any) -> Callable[[dict[Any, Any]], Any]:
    def inner(context: dict[Any, Any]) -> Any:
        return compute(argument, compute(context_sub, context))

    return inner


def field(key: str, *args) -> Callable[[dict[Any, Any] | list[Any]], Any]:
    def inner(context: dict[Any, Any] | list[Any]) -> Any:
        return (
            context.get(compute(key, context), compute(args[0], context))
            if args
            else itemgetter(compute(key, context))(context)
        )

    return inner


def value[T](value: T) -> Callable[[dict[Any, Any]], T]:
    def inner(_: Any) -> T:
        return value

    return inner



import operator
from collections.abc import Callable
from numbers import Number
from typing import Any

from genruler.common import binary


def add(*arguments: Any) -> Callable[[dict[Any, Any]], Number]:
    return binary(operator.add, arguments)


def subtract(*arguments: Any) -> Callable[[dict[Any, Any]], Number]:
    return binary(operator.sub, arguments)


def multiply(*arguments: Any) -> Callable[[dict[Any, Any]], Number]:
    return binary(operator.mul, arguments)


def divide(*arguments: Any) -> Callable[[dict[Any, Any]], Number]:
    return binary(operator.truediv, arguments)


def modulo(*arguments: Any) -> Callable[[dict[Any, Any]], Number]:
    return binary(operator.mod, arguments)



from dataclasses import dataclass
from funcparserlib.lexer import Token, make_tokenizer
from funcparserlib.parser import NoParseError, Parser, forward_decl, many, some


@dataclass
class Symbol:
    """Represents a symbol in the S-expression."""

    name: str

    def __str__(self) -> str:
        return self.name


def make_sexp_tokenizer():
    """Create a tokenizer for S-expressions.
    
    Returns:
        A tuple of (tokenizer function, set of useless token types)
    """
    specs = [
        ("SPACE", (r"[ \t\r\n]+",)),
        ("LPAREN", (r"\(",)),
        ("RPAREN", (r"\)",)),
        ("STRING", (r'"[^"]*"',)),
        ("NUMBER", (r"-?\d+\.\d+|-?\d+",)),
        (
            "SYMBOL",
            (r'[^\s\(\)"]+',),
        ),  # Matches any non-space, non-paren, non-quote characters
    ]
    useless = ["SPACE"]
    return make_tokenizer(specs), set(useless)


def make_parser() -> Parser:
    """Create a parser for S-expressions.
    
    Returns:
        A parser that converts tokens into an AST
    """
    # Forward declaration for recursive definitions
    expr = forward_decl()

    # Basic parsers for each token type
    def make_token_parser(token_type: str):
        return some(lambda t: t.type == token_type) >> (lambda t: t.value)

    lparen = make_token_parser("LPAREN")
    rparen = make_token_parser("RPAREN")

    # Convert numbers to actual Python numbers
    number = some(lambda t: t.type == "NUMBER") >> (
        lambda t: float(t.value) if "." in t.value else int(t.value)
    )

    # Remove quotes from strings
    string = some(lambda t: t.type == "STRING") >> (
        lambda t: t.value[1:-1]
    )  # Remove surrounding quotes

    # Convert symbols to Symbol objects
    symbol = some(lambda t: t.type == "SYMBOL") >> (lambda t: Symbol(t.value))

    # Atom can be number, string, or symbol
    atom = number | string | symbol

    # List is a sequence of expressions in parentheses
    list_expr = (lparen + many(expr) + rparen) >> (
        lambda x: x[1]
    )  # Extract items between parentheses

    # Only allow list expressions at the top level
    expr.define(atom | list_expr)
    top_level = (lparen + many(expr) + rparen) >> (
        lambda x: x[1]
    )  # Extract items between parentheses

    return top_level


def read(input: str) -> list:
    """Read an S-expression string into an AST.
    
    Args:
        input: The S-expression string to parse
        
    Returns:
        The parsed AST
        
    Raises:
        ValueError: If the input cannot be parsed
    """
    tokenizer, useless = make_sexp_tokenizer()
    try:
        # Tokenize input
        tokens = [
            Token(t.type, t.value) for t in tokenizer(input) if t.type not in useless
        ]

        # Parse tokens into AST
        parser = make_parser()
        return parser.parse(tokens)

    except NoParseError as e:
        raise ValueError(f"Parse error: {e}")



import operator
from collections.abc import Callable
from typing import Any, Literal

from genruler.common import binary
from genruler.library import compute


def and_(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.and_, arguments)


def contradiction() -> Callable[[dict[Any, Any]], Literal[False]]:
    def inner(_: Any) -> Literal[False]:
        return False

    return inner


def or_(*arguments: Any) -> Callable[[dict[Any, Any]], bool]:
    return binary(operator.or_, arguments)


def not_(argument: Any) -> Callable[[dict[Any, Any]], bool]:
    def inner(context: dict[Any, Any]) -> bool:
        return not compute(argument, context)

    return inner


def tautology() -> Callable[[dict[Any, Any]], Literal[True]]:
    def inner(_: Any) -> Literal[True]:
        return True

    return inner



from collections.abc import Callable
from functools import reduce
from typing import Any

from .library import compute


def binary[T](
    operation: Callable[[Any, Any], T], arguments: tuple[Any]
) -> Callable[[Any], T]:
    def inner(context: dict[Any, Any]) -> T:
        return reduce(operation, (compute(argument, context) for argument in arguments))

    return inner



from collections.abc import Callable
from typing import Any

import hy

from .lexer import read
from .library import evaluate, evaluate2


def parse(input: str, use_hy: bool = True) -> Callable[[Any], Any]:
    """Parse an S-expression string into a callable function.

    Args:
        input: The S-expression string to parse
        use_hy: If True, use Hy's reader (default). If False, use our custom reader.

    Returns:
        A callable function that takes a context argument

    Raises:
        ValueError: If the input cannot be parsed
        AssertionError: If the result is not callable
    """
    if use_hy:
        result = evaluate(hy.read(input))
    else:
        result = evaluate2(read(input))

    assert callable(result)
    return result


from typing import Any, Callable

from genruler.library import compute
from genruler.modules import basic


def field(key: str, *args) -> Callable[[dict[Any, Any] | list[Any]], Any]:
    def inner(context: dict[Any, Any] | list[Any]) -> Any:
        return str(compute(basic.field(compute(key, context), *args), context))

    return inner


def concat(link: Any, *arguments: Any) -> Callable[[dict[Any, Any]], str]:
    def inner(context: dict[Any, Any]) -> str:
        return compute(link, context).join(
            compute(argument, context) for argument in arguments
        )

    return inner


def concat_fields(link: Any, *arguments: Any) -> Callable[[dict[Any, Any]], str]:
    def inner(context: dict[Any, Any]) -> str:
        print(arguments)
        return concat(
            link,
            *[field(compute(argument, context))(context) for argument in arguments],
        )(context)

    return inner


def lower(argument: Any) -> Callable[[dict[Any, Any]], str]:
    def inner(context: dict[Any, Any]) -> str:
        return compute(argument, context).lower()

    return inner
